`include "internal_defines.vh"

//// memory_controller: Handles calculations for interfacing with memory
////
//// mem_data_in  (output) - Output to memory with data to write
//// mem_addr     (output) - Output to memory to select word to write
//// mem_write_en (output) - Output to memory to enable write
//// mc__out      (output) - Value read from memory
//// mem_data_out (input)  - Input from the memory with the read data
//// mc__address  (input)  - The naturally aligned address of the access
//// mc__size     (input)  - The size of the access
//// mc__in       (input)  - The data to write
//// mc__we       (input)  - Whether this is a write operation
//// mc__se       (input)  - Whether to sign extend byte/halfword reads
////
module memory_controller(/*AUTOARG*/
   // Outputs
   mem_data_in, mem_addr, mem_write_en, mc__out, mc__AdEL_data,
   mc__AdES, mc__DBE,
   // Inputs
   mem_data_out, mem_excpt, mc__address, mc__size, mc__in, mc__we,
   mc__se
   );

   output reg [31:0] mem_data_in;
   output [29:0] mem_addr;
   output reg [3:0]  mem_write_en;
   output reg [31:0] mc__out;
   output reg        mc__AdEL_data;
   output reg        mc__AdES;
   output reg        mc__DBE;
   input [31:0]      mem_data_out;
   input 	     mem_excpt;
   input [31:0]      mc__address;
   input [1:0] 	     mc__size;
   input [31:0]      mc__in;
   input 	     mc__we;
   input 	     mc__se;

   assign mem_addr = mc__address[31:2];

   always @(*) begin
      mc__AdEL_data = 1'b0;
      mc__AdES = 1'b0;
      mc__DBE = mc__size == `MC_DISABLED ? 1'b0 : mem_excpt;
      mc__out = 32'bx;
      mem_data_in = 32'bx;
      mem_write_en = 4'b0;

      case (mc__size)
	`MC_BYTE: begin
	   if (mc__we) begin
	      mem_write_en = 4'b1000 >> mc__address[1:0];
	      mem_data_in = mc__in[7:0] << (8 * mc__address[1:0]);
	   end else begin
	      mc__out = (mem_data_out >> (8 * mc__address[1:0])) & 8'hff;
	      if (mc__se)
		mc__out = { {24{mc__out[7]}}, mc__out[7:0] };
	   end
	end

	`MC_HALFWORD: begin
	   if (mc__we) begin
	      if (mc__address[0])
		mc__AdES = 1'b1;

	      mem_write_en = 4'b1100 >> (2 * mc__address[1]);
	      mem_data_in = mc__in[15:0] << (16 * mc__address[1]);
	   end else begin
	      if (mc__address[0])
		mc__AdEL_data = 1'b1;

	      mc__out = (mem_data_out >> (16 * mc__address[1])) & 16'hffff;
	      if (mc__se)
		mc__out = { {16{mc__out[15]}}, mc__out[15:0] };
	   end
	end

	`MC_WORD: begin
	   if (mc__we) begin
	      if (mc__address[1:0])
		mc__AdES = 1'b1;
	      
	      mem_write_en = 4'b1111;
	      mem_data_in = mc__in;
	   end else begin
	      if (mc__address[1:0])
		mc__AdEL_data = 1'b1;

	      mc__out = mem_data_out;
	   end
	end
      endcase // case (mc__size)
   end
endmodule
